//****************************************************************************
// *                                                                         *
// *           Aflow STEFANO CURTAROLO - Duke University 2003-2018           *
// *                  Marco Esters - Duke University 2018                    *
// *                                                                         *
//****************************************************************************
//
// This class implements n-dimensional tensors using a 1D array mimicking a
// tensor. See aurostd_xtensor.h for descriptions of the class attributes.

#ifndef _AUROSTD_XTENSOR_CPP_
#define _AUROSTD_XTENSOR_CPP_

#ifndef _AUROSTD_XSCALAR_H_
#include "aurostd_xscalar.h"
#endif

#ifndef _AUROSTD_XVECTOR_H_
#include "aurostd_xvector.h"
#endif

#ifndef _AUROSTD_XMATRIX_H_
#include "aurostd_xmatrix.h"
#endif

#ifndef _AUROSTD_XTENSOR_H_
#include "aurostd_xtensor.h"
#endif

#ifndef _AUROSTD_XCOMBOS_H_
#include "aurostd_xcombos.h"
#endif

#ifndef _AUROSTD_XERROR_H_
#include "aurostd_xerror.h"
#endif

#define _DEBUG_XTENSOR_ false

static const std::string _SUBTENSOR_ERR_PREFIX_ = "aurostd::_subtensor::";
static const std::string _XTENSOR_ERR_PREFIX_ = "aurostd::xtensor::";

using aurostd::xerror;
using aurostd::xcombos;

/********************************* SUBTENSOR ********************************/
// Helper class for indexing of xtensor. Indexing xtensor is performed by
// storing the indices subsequently into _subtensor and then by using the
// assignment operator or type casting (see conversion operators).
//
// The indices in _subtensor must be 0-based, so they need to be converted
// before being passed into the constructor or appended to the indices vector.
namespace aurostd{

//Constructors////////////////////////////////////////////////////////////////
template<class utype>
_subtensor<utype>::_subtensor(const int& i, const std::vector<int>& shf,
                              const std::vector<int>& shp, const std::vector<int>& uind,
                              const std::vector<int>& lind, utype* crp) {
  indices.resize(1, i);
  uindex = uind;
  lindex = lind;
  shifts = shf;
  shape = shp;
  corpus = crp;
}

template<class utype>
_subtensor<utype>::_subtensor(const std::vector<int>& ind, const std::vector<int>& shf,
                              const std::vector<int>& shp, const std::vector<int>& uind,
                              const std::vector<int>& lind, utype* crp) {
  indices = ind;
  uindex = uind;
  lindex = lind;
  shifts = shf;
  shape = shp;
  corpus = crp;
}

//Destructor//////////////////////////////////////////////////////////////////
template<class utype>
_subtensor<utype>::~_subtensor() {
  indices.clear();
}

//Indexing operator///////////////////////////////////////////////////////////
template<class utype>
_subtensor<utype> _subtensor<utype>::operator[] (int i) {
  if ((i >= lindex[indices.size()]) && (indices.size() < shape.size()) && (i <= uindex[indices.size()])) {
    i -= lindex[indices.size()];
    indices.push_back(i);
    return *this;
  // Exception handling
  } else {
    std::string function = _SUBTENSOR_ERR_PREFIX_ + "operator[]";
    std::stringstream message;
    if (indices.size() == shape.size()) {
       message << "Cannot subscribe tensor any further (tensor size: " << shape.size() << ").";
    } else {
       message << "Index " << i << " out of bounds for dimension " << (indices.size());
       message << " (lindex = " << lindex[indices.size()] <<", uindex = " << uindex[indices.size()] << ")";
    }
    throw xerror(function, message, _INDEX_BOUNDS_);
  }
}

template<class utype>
_subtensor<utype> _subtensor<utype>::operator() (std::vector<int> ind) {
  // Exception handling
  std::string function = _SUBTENSOR_ERR_PREFIX_ + "operator()";
  std::stringstream message;
  if (indices.size() + ind.size() > shape.size()) {
    message  << "Too many indices for tensor with " << shape.size() << " dimensions";
    message  << " (" << (indices.size() + ind.size()) << " provided).";
    throw xerror(function, message, _INDEX_MISMATCH_);
  }
  for (uint i = 0; i < ind.size(); i++) {
    if ((ind[i] >= lindex[indices.size()]) && (ind[i] <= uindex[indices.size()])) {
      indices.push_back(ind[i] - lindex[indices.size()]);
    } else {
       message << "Index " << ind[i] << " out of bounds for dimension " << (indices.size());
       message << " (lindex = " << lindex[indices.size()] <<", uindex = " << uindex[indices.size()] << ")";
       throw xerror(function, message, _INDEX_BOUNDS_);
    }
  }
  // All checks passed
  return *this;
}

// BEGIN Operators
//Unary operators//////////////////////////////////////////////////////////////
//// BEGIN with scalars
template<class utype>
_subtensor<utype>& _subtensor<utype>::operator+=(utype scalar) {
  int shift = getShift();
  if (indices.size() == shape.size()) {
    corpus[shift] += scalar;
  } else {
    int n = 1;
    for (uint i = indices.size(); i < shape.size(); i++) {
      n *= shape[i];
    }
    for (int s = 0; s < n; s++) {
      corpus[shift+s] += scalar;
    }
  }
  return *this;
}

template<class utype>
_subtensor<utype>& _subtensor<utype>::operator-=(utype scalar) {
  int shift = getShift();
  if (indices.size() == shape.size()) {
    corpus[shift] -= scalar;
  } else {
    int n = 1;
    for (uint i = indices.size(); i < shape.size(); i++) {
      n *= shape[i];
    }
    for (int s = 0; s < n; s++) {
      corpus[shift+s] -= scalar;
    }
  }
  return *this;
}

template<class utype>
_subtensor<utype>& _subtensor<utype>::operator*=(utype scalar) {
  int shift = getShift();
  if (indices.size() == shape.size()) {
    corpus[shift] *= scalar;
  } else {
    int n = 1;
    for (uint i = indices.size(); i < shape.size(); i++) {
      n *= shape[i];
    }
    for (int s = 0; s < n; s++) {
      corpus[shift+s] *= scalar;
    }
  }
  return *this;
}

template<class utype>
_subtensor<utype>& _subtensor<utype>::operator/=(utype scalar) {
  int shift = getShift();
  if (indices.size() == shape.size()) {
    corpus[shift] /= scalar;
  } else {
    int n = 1;
    for (uint i = indices.size(); i < shape.size(); i++) {
      n *= shape[i];
    }
    for (int s = 0; s < n; s++) {
      corpus[shift+s] /= scalar;
    }
  }
  return *this;
}
//// END with scalars

//// BEGIN with tensors
template<class utype>
_subtensor<utype>& _subtensor<utype>::operator+=(const _subtensor<utype>& st) {
  int shift = getShift();
  int st_shift = st.getShift();
  if (sameShape(st)) {
    if (indices.size() == shape.size()) {
      corpus[shift] += st.corpus[st_shift];
    } else {
      int n = 1;
      for (uint i = indices.size(); i < shape.size(); i++) {
        n *= shape[i];
      }
      for (int s = 0; s < n; s++) {
        corpus[shift+s] += st.corpus[st_shift+s];
      }
    }
  } else {
    std::string function = _SUBTENSOR_ERR_PREFIX_ + "operator+=";
    std::string message = "Subtensors have different shapes.";
    throw xerror(function, message, _INDEX_MISMATCH_);
  }
  return *this;
}

template<class utype>
_subtensor<utype>& _subtensor<utype>::operator-=(const _subtensor<utype>& st) {
  int shift = getShift();
  int st_shift = st.getShift();
  if (sameShape(st)) {
    if (indices.size() == shape.size()) {
      corpus[shift] -= st.corpus[st_shift];
    } else {
      int n = 1;
      for (uint i = indices.size(); i < shape.size(); i++) {
        n *= shape[i];
      }
      for (int s = 0; s < n; s++) {
        corpus[shift+s] -= st.corpus[st_shift+s];
      }
    }
  } else {
    std::string function = _SUBTENSOR_ERR_PREFIX_ + "operator+=";
    std::string message = "Subtensors have different shapes.";
    throw xerror(function, message, _INDEX_MISMATCH_);
  }
  return *this;
}

template<class utype>
_subtensor<utype>& _subtensor<utype>::operator+=(const xtensor<utype>& tensor) {
  int shift = getShift();
  if (sameShape(tensor)) {
    if (indices.size() == shape.size()) {
      corpus[shift] += tensor[0];
    } else {
      int s = 0;
      xcombos ind(tensor.shape, 'E');
      while (ind.increment()) {
        std::vector<int> indices = ind.getCombo();
        for (uint i = 0; i < indices.size(); i++) {
          indices[i] += tensor.lindex[i];
        }
        corpus[shift+s] += tensor(indices);
        s++;
      }
    }
  } else {
    std::string function = _SUBTENSOR_ERR_PREFIX_ + "operator+=";
    std::string message = "Subtensor and tensor have different shapes.";
    throw xerror(function, message, _INDEX_MISMATCH_);
  }
  return *this;
}

template<class utype>
_subtensor<utype>& _subtensor<utype>::operator-=(const xtensor<utype>& tensor) {
  int shift = getShift();
  if (sameShape(tensor)) {
    if (indices.size() == shape.size()) {
      corpus[shift] -= tensor[0];
    } else {
      int s = 0;
      xcombos ind(tensor.shape, 'E');
      while (ind.increment()) {
        std::vector<int> indices = ind.getCombo();
        for (uint i = 0; i < indices.size(); i++) {
          indices[i] += tensor.lindex[i];
        }
        corpus[shift+s] -= tensor(indices);
        s++;
      }
    }
  } else {
    std::string function = _SUBTENSOR_ERR_PREFIX_ + "operator+=";
    std::string message = "Subtensor and tensor have different shapes.";
    throw xerror(function, message, _INDEX_MISMATCH_);
  }
  return *this;
}
//// END with tensors

//Conversion operators////////////////////////////////////////////////////////
// Converts a _subtensor into utype to be assigned to a utype variable. For
// the conversion of _subtensor into xtensor, see the (copy) constructor for
// xtensor.
template<class utype>
_subtensor<utype>::operator utype() const {
  int shift = getShift();
  return *(corpus + shift);
}

// Assigns a utype to an xtensor element
template<class utype>
_subtensor<utype>& _subtensor<utype>::operator=(const utype& val) {
  int shift = getShift();
  if (indices.size() == shape.size()) {
    corpus[shift] = val;
  // Exception handling
  } else {
    std::string function = _XTENSOR_ERR_PREFIX_ + "operator=";
    std::stringstream message;
    message << "Cannot assign single value to tensor of size " << (shape.size() - indices.size()) << ".";
    throw xerror(function, message, _INDEX_MISMATCH_);
  }
  return *this;
}

// Assigns a subtensor to another subtensor
template<class utype>
_subtensor<utype>& _subtensor<utype>::operator=(const _subtensor<utype>& st) {
  if (sameShape(st)) {
    int start = getShift();
    int start_st = st.getShift();
    int size = shifts[indices.size() - 1];
    for (int i = 0; i < size; i++) {
      corpus[start+i] = st.corpus[start_st+i];
    }
  } else {
    std::string function = _SUBTENSOR_ERR_PREFIX_ + "operator=";
    std::string message = "Subtensors have different shapes.";
    throw xerror(function, message, _INDEX_MISMATCH_);
  }
  return *this;
}

// Assigns a tensor to a subtensor
template<class utype>
_subtensor<utype>& _subtensor<utype>::operator=(const xtensor<utype>& tensor) {
  if (sameShape(tensor)) {
    int st_index = getShift();
    xcombos combos(tensor.shape, 'E');
    while (combos.increment()) {
      std::vector<int> indices = shiftIndices(combos.getCombo(), tensor.lindex);
      corpus[st_index] = tensor(indices);
      st_index++;
    }
  } else {
    std::string function = _SUBTENSOR_ERR_PREFIX_ + "operator=";
    std::string message = "Subtensor and tensor have different shapes.";
    throw xerror(function, message, _INDEX_MISMATCH_);
  }
  return *this;
}

// Assigns a vector to a 1D slice of a tensor
template<class utype>
_subtensor<utype>& _subtensor<utype>::operator=(const std::vector<utype>& vec) {
  bool throw_exception = false;
  std::stringstream message;
  // Slice must be 1D
  if (indices.size() == (shape.size() - 1)) {
    // Tensor slice and vector must have same size
    if ((int)vec.size() == shape[indices.size()]) {
      int shift = getShift();
      for (uint i = 0; i < vec.size(); i++) {
        corpus[shift+i] = vec[i];
      }
    // Exception handling.
    } else {
      throw_exception = true;
      message << "Tensor slice and vector have different sizes. ";
      message << "Size of tensor slice: " << shape[indices.size()] << "; ";
      message << "size of vector: " << vec.size() << ".";
    }
  } else {
    throw_exception = true;
    message << "Tensor slice must be 1D to assign vector (is " << (shape.size() - indices.size()) << "D).";
  }
  if (throw_exception) {
    std::string function = _SUBTENSOR_ERR_PREFIX_ + "operator=";
    throw xerror(function, message, _INDEX_MISMATCH_);
  }
  return *this;
}

// Assigns an xvector to a 1D slice of a tensor
template<class utype>
_subtensor<utype>& _subtensor<utype>::operator=(const aurostd::xvector<utype>& xvec) {
  bool throw_exception = false;
  std::stringstream message;
  // Tensor slice must be 1D
  if (indices.size() == (shape.size() - 1)) {
    // Tensor slice and xvector must have same size
    if (xvec.rows == shape[indices.size()]) {
      int st_shift = getShift();
      for (int i = 0; i < xvec.rows; i++) {
        corpus[st_shift+i] = xvec(xvec.lrows+i);
      }
    // Exception handling
    } else {
      throw_exception = true;
      message << "Tensor slice and xvector have different sizes. ";
      message << "Size of tensor slice: " << shape[indices.size()] << "; ";
      message << "size of xvector: " << xvec.rows << ".";
    }
  } else {
    throw_exception = true;
    message << "Tensor slice must be 1D to assign xvector (is " << (shape.size() - indices.size()) << "D).";
  }
  if (throw_exception) {
    std::string function = _SUBTENSOR_ERR_PREFIX_ + "operator=";
    throw xerror(function, message, _INDEX_MISMATCH_);
  }
  return *this;
}

// Assigns an xmatrix to a 2D slice of a tensor
template<class utype>
_subtensor<utype>& _subtensor<utype>::operator=(const aurostd::xmatrix<utype>& xmat) {
  bool throw_exception = false;
  std::stringstream message;
  // Tensor slice must be 2D
  if (indices.size() == (shape.size() - 2)) {
    // Tensor slice must have same shape as matrix
    if ((xmat.rows == shape[indices.size()]) && (xmat.cols == shape[indices.size() + 1])) {
      int st_index = getShift();
      for (int i = 0; i < xmat.rows; i++) {
        for (int j = 0; j < xmat.cols; j++) {
          corpus[st_index] = xmat(xmat.lrows+i, xmat.lcols+j);
          st_index++;
        }
      }
    } else {
      throw_exception = true;
      message << "Tensor slice and xmatrix have different shapes. ";
      message << "Shape of tensor slice: (" << shape[indices.size()] << ", " << shape[indices.size() + 1] << "); ";
      message << "shape of xmatrix: (" << xmat.rows << ", " << xmat.cols << ").";
    }
  } else {
    throw_exception = true;
    message << "Tensor slice must be 2D to assign xmatrix (is " << (shape.size() - indices.size()) << "D).";
  }
  if (throw_exception) {
    std::string function = _SUBTENSOR_ERR_PREFIX_ + "operator=";
    throw xerror(function, message, _INDEX_MISMATCH_);
  }
  return *this;
}

// END Operators

// BEGIN Helper functions
//getShift////////////////////////////////////////////////////////////////////
template<class utype>
int _subtensor<utype>::getShift() const {
  int shift = 0;
  for (uint i = 0; i < indices.size(); i++) {
    shift += indices[i] * shifts[i];
  }
  return shift;
}

//sameShape///////////////////////////////////////////////////////////////////
template<class utype>
bool _subtensor<utype>::sameShape(const _subtensor<utype>& st) {
  uint size1 = shape.size() - indices.size();
  uint size2 = st.shape.size() - st.indices.size();
  if (size1 != size2) {
    return false;
  } else {
    int diff = (int) (size2 - size1);
    for (uint i = indices.size(); i < shape.size(); i++) {
      if (shape[i] != st.shape[i+diff]) {
        return false;
      }
    } 
  }
  return true;
}

template<class utype>
bool _subtensor<utype>::sameShape(const xtensor<utype>& tensor) {
  uint size = shape.size() - indices.size();
  if (size != tensor.ndim) {
    return false;
  } else {
    for (uint i = 0; i < size; i++) {
      if (shape[i+indices.size()] != tensor.shape[i]) {
        return false;
      }
    }
  }
  return true;
}
// END Helper functions
} // namespace aurostd

/******************************* END SUBTENSOR ******************************/

/********************************** XTENSOR *********************************/
// The xtensor class. As with xvector and xmatrix, xtensors are initialized
// upper indices (uind/uindex) and the lower indices (lind/lindex), and the
// default indexing is 1-based. uindex and lindex can be provided using
// (x)vectors or an integer describing the number of dimensions and the lower
// index for all dimensions (default 1).
namespace aurostd {

// BEGIN Constructors/Destructors
//Constructors////////////////////////////////////////////////////////////////
template<class utype>
xtensor<utype>::xtensor() {  // Default constructor
  std::vector<int> uind(3, 3);
  std::vector<int> lind(3, 1);
  buildTensor(uind, lind);
}

template<class utype>
xtensor<utype>::xtensor(const std::vector<int>& uind) {
  std::vector<int> lind(uind.size(), 1);
  buildTensor(uind, lind);
}

template<class utype>
xtensor<utype>::xtensor(const std::vector<int>& uind, const std::vector<int>& lind) {
  buildTensor(uind, lind);
}


template<class utype>
xtensor<utype>::xtensor(const aurostd::xvector<int>& xuind) {
  std::vector<int> uind = aurostd::xvector2vector(xuind);
  std::vector<int> lind(uind.size(), 1);
  buildTensor(uind, lind);
}

template<class utype>
xtensor<utype>::xtensor(const aurostd::xvector<int>& xuind, const aurostd::xvector<int>& xlind) {
  std::vector<int> uind = aurostd::xvector2vector(xuind);
  std::vector<int> lind = aurostd::xvector2vector(xlind);
  buildTensor(uind, lind);
}

template<class utype>
xtensor<utype>::xtensor(const int& ui, const int& li) {
// Creates a tensor with (ui-li + 1) dimensions of size (ui-li). For example,
// a 3x3x3 tensor would be xtensor(3) (1-based) or xtensor(2, 0) (0-based).
  int n = ui - li + 1;
  std::vector<int> uind(n, n);
  std::vector<int> lind(n, li);
  buildTensor(uind, lind);
}

template<class utype>
xtensor<utype>::xtensor(const _subtensor<utype>& st) {
  std::vector<int> uind, lind;
  if (st.indices.size() == st.shape.size()) {
    uind.resize(1, 1);
    lind.resize(1, st.lindex[st.shape.size()-1]);
  } else {
    for (uint i = st.indices.size(); i < st.shape.size(); i++) {
      uind.push_back(st.uindex[i]);
      lind.push_back(st.lindex[i]);
    }
  }
  buildTensor(uind, lind);
  int start = st.getShift();
  int end = start + st.shifts[st.indices.size() - 1];
  for (int i = start; i < end; i++) {
    corpus[i - start] = st.corpus[i];
  }
}

//buildTensor/////////////////////////////////////////////////////////////////
template<class utype>
void xtensor<utype>::buildTensor(const std::vector<int>& uind, const std::vector<int>& lind) {
  bool LDEBUG = (false || XHOST.DEBUG || _DEBUG_XTENSOR_);
  if (checkInit(uind, lind)) {
    uindex = uind;
    lindex = lind;
    ndim = uindex.size();
    shape.resize(uind.size());
    for (uint i = 0; i < shape.size(); i++) {
      shape[i] = uindex[i] - lindex[i] + 1;
    }
    if (LDEBUG) {
      std::cerr << _XTENSOR_ERR_PREFIX_ << "buildTensor - Tensor dimensions:";
      for (uint i = 0; i < shape.size(); i++) {
        std::cerr << " " << shape[i];
      }
      std::cerr << std::endl;
    }
    if ((uind.size() == 1) && ((uindex[0] - lindex[0]) == 1)) {
      ndim  = 0;
      shifts.resize(1, 1);
    } else {
      ndim = shape.size();
      shifts.resize(ndim, 1);
    }
    for (uint s = 0; s < ndim - 1; s++) {
      for (uint i = s + 1; i < ndim; i++) {
        shifts[s] *= shape[i];
      }
    }
    if (ndim < 2) {
      is_cubic = false;
    } else {
      for (uint i = 1; i < ndim; i++) {
        if (shape[i] != shape[i-1]) {
          is_cubic = false;
          i = ndim;
        }
      }
    }
    size = (char) (sizeof(utype));
    nelements = 1;
    for (uint i = 0; i < ndim; i++) {
      nelements *= shape[i];
    }
    tsize = (long int) size * nelements;
    corpus = new utype [nelements];
    for (int i = 0; i < nelements; i++) {
      corpus[i] = (utype) 0;
    }
    if (LDEBUG) {
      std::cerr << "is_cubic = " << ((is_cubic)? "true" : "false")  << ", sizeof = " << size << ", tsize = " << tsize << std::endl;
    }
    if (!corpus) {
      std::string function = _XTENSOR_ERR_PREFIX_ + "xtensor";
      std::string message = "Allocation failure in xtensor constructor.";
      throw xerror(function, message, _ALLOC_ALLOCATE_);
    }
  // Exception handling
  } else {
    tsize = 0;  // For destructor
    std::string function = _XTENSOR_ERR_PREFIX_ + "xtensor";
    std::stringstream message;
    int code;
    if (uind.size() != lind.size()) {
      message << "uindex and lindex are of different size.";
      code = _INDEX_MISMATCH_;
    } else {
      if (uind.size() == 0 || lind.size() == 0) {
        message << "Cannot initialize tensor with empty vector.";
      } else {
        message << "Vector describing tensor indices contains illegal values (uindex < lindex).";
      }
      code = _INDEX_ILLEGAL_;
    }
    throw xerror(function, message, code);
  }
}

//Copy Constructors///////////////////////////////////////////////////////////
template<class utype>
xtensor<utype>::xtensor(const xtensor<utype>& that) {
  is_cubic = that.is_cubic;
  lindex = that.lindex;
  nelements = that.nelements;
  ndim = that.ndim;
  shape = that.shape;
  size = that.size;
  tsize = that.tsize;
  shifts = that.shifts;
  uindex = that.uindex;
  corpus = new utype [nelements];
  for (int i = 0; i < nelements; i++) {
    corpus[i] = that.corpus[i];
  }
}

template<class utype>
xtensor<utype> xtensor<utype>::operator=(const xtensor<utype>& that) {
  if (corpus != that.corpus) {
    // if shapes are different, build a new xtensor
    bool destroy = false;
    if (ndim != that.ndim) {
      destroy = true;
    } else {
      for (uint d = 0; d < ndim; d++) {
        if (shape[d] != that.shape[d]) {
          destroy = true;
          d = ndim;
        }
      }
    }
    if (destroy) {
      delete [] (corpus);
      corpus = new utype [that.nelements];
      is_cubic = that.is_cubic;
      ndim = that.ndim;
      nelements = that.nelements;
      shape = that.shape;
      shifts = that.shifts;
      size = that.size;
      tsize = that.tsize;
    }
    lindex = that.lindex;
    uindex = that.uindex;
    for (int i = 0; i < nelements; i++) {
      corpus[i] = that.corpus[i];
    }
  }
  return *this;
}

template<class utype>
xtensor<utype> xtensor<utype>::operator=(const _subtensor<utype>& st) {
  return xtensor<utype>(st);
}

//Destructor//////////////////////////////////////////////////////////////////
template<class utype>
xtensor<utype>::~xtensor() {
  if (tsize > 0) {
    delete [] (corpus);
  }
  shape.clear();
  uindex.clear();
  lindex.clear();
}
// END Constructors/Destructors

// BEGIN helper functions
//checkInit///////////////////////////////////////////////////////////////////
// helper function for the constructor to test the validity of the input std::vector
template<class utype>
bool xtensor<utype>::checkInit(const std::vector<int>& uind, const std::vector<int>& lind) {
  if (uind.size() != lind.size()) {
    return false;
  }
  if (uind.size() == 0 || lind.size() == 0) {
    return false;
  }
  for (uint i = 0; i < uind.size(); i++) {
    if (uind[i] < lind[i]) {
      return false;
    }
  }
  return true;
}

//sameShape///////////////////////////////////////////////////////////////////
// helper function for operations between two tensors to check if both have
// the same shape
template<class utype>
bool xtensor<utype>::sameShape(const xtensor<utype>& tensor) const {
  for (uint d = 0; d < ndim; d++) {
    if (shape[d] != tensor.shape[d]) {
      return false;
    }
  }
  return true;
}

template<class utype>
bool xtensor<utype>::sameShape(const _subtensor<utype>& st) const {
  uint size = st.shape.size() - st.indices.size();
  if (size != ndim) {
    return false;
  } else {
    for (uint i = 0; i < size; i++) {
      if (st.shape[i+st.indices.size()] != shape[i]) {
        return false;
      }
    }
  }
  return true;
}
// END helper functions

// BEGIN operators
// BEGIN index operators
template<class utype>
_subtensor<utype> xtensor<utype>::operator()(std::vector<int> indices) const {
  // Exception handling
  std::string function = _XTENSOR_ERR_PREFIX_ + "operator()";
  std::stringstream message;
  if (indices.size() > ndim) {
    message  << "Too many indices for tensor with " << ndim << " dimensions (" << indices.size() << " provided).";
    throw xerror(function, message, _INDEX_MISMATCH_);
  } else {
    for (uint i = 0; i < indices.size(); i++) {
      if ((indices[i] < lindex[i]) || indices[i] > uindex[i]) {
        message << "Index " << indices[i] << " out of bounds for dimension " << i;
        message << " (lindex = " << lindex[i] << ", uindex = " << uindex[i] <<").";
        throw xerror(function, message, _INDEX_BOUNDS_);
      } else {
        indices[i] -= lindex[i];
      }
    }
  }
  // All checks passed
  return _subtensor<utype>(indices, shifts, shape, uindex, lindex, corpus);
}

template<class utype>
_subtensor<utype> xtensor<utype>::operator()(aurostd::xvector<int> xvec) const {
  std::vector<int> indices = aurostd::xvector2vector(xvec);
  // Exception handling
  std::string function = _XTENSOR_ERR_PREFIX_ + "operator()";
  std::stringstream message;
  if (indices.size() > ndim) {
    message << "Too many indices for tensor with " << ndim << " dimensions (" << indices.size() << " provided).";
    throw xerror(function, message, _INDEX_MISMATCH_);
  } else {
    for (uint i = 0; i < ndim; i++) {
      if (indices[i] < lindex[0] || indices[i] > uindex[i]) {
        message << "Index " << indices[i] << " out of bounds for dimension "<< i;
        message << " (lindex = " << lindex[i] << ", uindex = " << uindex[i] << ").";
        throw xerror(function, message, _INDEX_BOUNDS_);
      } else {
        indices[i] -= lindex[i];
      }
    }
  }
  // All checks passed
  return _subtensor<utype>(indices, shifts, shape, uindex, lindex, corpus);
}

template<class utype>
_subtensor<utype> xtensor<utype>::operator[](int i) const {
  if ((i >= lindex[0]) && (i <= uindex[0])) {
    i -= lindex[0];
    return _subtensor<utype>(i, shifts, shape, uindex, lindex, corpus);
  // Exception handling
  } else {
    std::string function = _XTENSOR_ERR_PREFIX_ + "operator[]";
    std::stringstream message;
    message << "Index " << i << " out of bounds for dimension 0";
    message << " (lindex = " << lindex[0] << ", uindex = " << uindex[0] << ").";
    throw xerror(function, message, _INDEX_BOUNDS_);
  }
}
// END index operators

//// BEGIN unary operators
////// BEGIN with scalars  
template<class utype>
xtensor<utype>& xtensor<utype>::operator+=(utype scalar) {
  for (int i = 0; i < nelements; i++) {
    corpus[i] += scalar;
  }
  return *this;
}

template<class utype>
xtensor<utype>& xtensor<utype>::operator-=(utype scalar) {
  for (int i = 0; i < nelements; i++) {
    corpus[i] -= scalar;
  }
  return *this;
}

template<class utype>
xtensor<utype>& xtensor<utype>::operator*=(utype scalar) {
  for (int i = 0; i < nelements; i++) {
    corpus[i] *= scalar;
  }
  return *this;
}

template<class utype>
xtensor<utype>& xtensor<utype>::operator/=(utype scalar) {
  for (int i = 0; i < nelements; i++) {
    corpus[i] /= scalar;
  }
  return *this;
}
////// END with scalars

////// BEGIN with xtensors
template<class utype>
xtensor<utype>& xtensor<utype>::operator+=(const xtensor<utype>& tensor) {
  bool LDEBUG = (false || XHOST.DEBUG || _DEBUG_XTENSOR_);
  if (LDEBUG) {
    std::cerr << _XTENSOR_ERR_PREFIX_ << "operator+= : Dimensions tensor 1:";
    for (uint i = 0; i < shape.size(); i++) {
      std::cerr << " " << shape[i];
    }
    std::cerr << ", dimensions tensor 2:";
    for (uint i = 0; i < tensor.shape.size(); i++) {
      std::cerr << " " << tensor.shape[i];
    }
    std::cerr << std::endl;
  }
  if (sameShape(tensor)) {
    xcombos combos(tensor.shape, 'E');
    int corpus_index = 0;
    while (combos.increment()) {
      std::vector<int> indices = shiftIndices(combos.getCombo(), lindex);
      corpus[corpus_index] += (utype) tensor(indices);
      corpus_index++;
    }
    return *this;
  // Exception handling
  } else {
    std::string function = _XTENSOR_ERR_PREFIX_ + "operator+=";
    std::string message = "Tensors are of different size.";
    throw xerror(function, message, _INDEX_MISMATCH_);
  }
}

template<class utype>
xtensor<utype>& xtensor<utype>::operator-=(const xtensor<utype>& tensor) {
  bool LDEBUG = (false || XHOST.DEBUG || _DEBUG_XTENSOR_);
  if (LDEBUG) {
    std::cerr << _XTENSOR_ERR_PREFIX_ << "operator-= : Dimensions tensor 1:";
    for (uint i = 0; i < shape.size(); i++) {
      std::cerr << " " << shape[i];
    }
    std::cerr << ", dimensions tensor 2:";
    for (uint i = 0; i < tensor.shape.size(); i++) {
      std::cerr << " " << tensor.shape[i];
    }
    std::cerr << std::endl;
  }
  if (sameShape(tensor)) {
    xcombos combos(tensor.shape, 'E');
    int corpus_index = 0;
    while (combos.increment()) {
      std::vector<int> indices = shiftIndices(combos.getCombo(), tensor.lindex);
      corpus[corpus_index] -= (utype) tensor(indices);
      corpus_index++;
    }
    return *this;
  // Exception handling
  } else {
    std::string function = _XTENSOR_ERR_PREFIX_ + "operator-=";
    std::string message = "Tensors are of different size.";
    throw xerror(function, message, _INDEX_MISMATCH_);
  }
}

template<class utype>
xtensor<utype> operator+(const xtensor<utype>& tensor) {
  return tensor;
}

template<class utype>
xtensor<utype> operator-(xtensor<utype> tensor) {
  xcombos combos(tensor.shape, 'E');
  while (combos.increment()) {
    std::vector<int> indices = shiftIndices(combos.getCombo(), tensor.lindex);
    tensor(indices) = -(utype) tensor(indices);
  }
  return tensor;
}
////// END with xtensors

////// BEGIN with _subtensors
template<class utype>
xtensor<utype>& xtensor<utype>::operator+=(const _subtensor<utype>& st) {
  bool LDEBUG = (false || XHOST.DEBUG || _DEBUG_XTENSOR_);
  if (LDEBUG) {
    std::cerr << _XTENSOR_ERR_PREFIX_ << "operator-= : Dimensions tensor:";
    for (uint i = 0; i < shape.size(); i++) {
      std::cerr << " " << shape[i];
    }
    std::cerr << ", dimensions subtensor:";
    if (st.indices.size() == st.shape.size()) {
      std::cerr << " 1" << std::endl;
    } else {
      for (uint i = st.indices.size(); i < st.shape.size(); i++) {
        std::cerr << " " << st.shape[i];
      }
      std::cerr << std::endl;
    }
  }
  if (sameShape(st)) {
    int shift = st.getShift();
    if (st.indices.size() == st.shape.size()) {
      corpus[0] += st.corpus[shift];
    } else {
      int n = 1;
      for (uint i = st.indices.size(); i < st.shape.size(); i++) {
        n *= st.shape[i];
      }
      for (int s = 0; s < n; s++) {
        corpus[s] += st.corpus[shift+s];
      }
    }
    return *this;
  } else {
    std::string function = _XTENSOR_ERR_PREFIX_ + "operator+=";
    std::string message = "Tensor and subtensor are of different size.";
    throw xerror(function, message, _INDEX_MISMATCH_);
  }
}

template<class utype>
xtensor<utype>& xtensor<utype>::operator-=(const _subtensor<utype>& st) {
  bool LDEBUG = (false || XHOST.DEBUG || _DEBUG_XTENSOR_);
  if (LDEBUG) {
    std::cerr << _XTENSOR_ERR_PREFIX_ << "operator-= : Dimensions tensor:";
    for (uint i = 0; i < shape.size(); i++) {
      std::cerr << " " << shape[i];
    }
    std::cerr << ", dimensions subtensor:";
    if (st.indices.size() == st.shape.size()) {
      std::cerr << " 1" << std::endl;
    } else {
      for (uint i = st.indices.size(); i < st.shape.size(); i++) {
        std::cerr << " " << st.shape[i];
      }
      std::cerr << std::endl;
    }
  }
  if (sameShape(st)) {
    int shift = st.getShift();
    if (st.indices.size() == st.shape.size()) {
      corpus[0] -= st.corpus[shift];
    } else {
      int n = 1;
      for (uint i = st.indices.size(); i < st.shape.size(); i++) {
        n *= st.shape[i];
      }
      for (int s = 0; s < n; s++) {
        corpus[s] -= st.corpus[shift+s];
      }
    }
    return *this;
  } else {
    std::string function = _XTENSOR_ERR_PREFIX_ + "operator+=";
    std::string message = "Tensor and subtensor are of different size.";
    throw xerror(function, message, _INDEX_MISMATCH_);
  }
}
////// END with _subtensors
//// END unary operators

//// BEGIN binary operators
////// BEGIN with scalars
template<class utype>
xtensor<utype> operator*(xtensor<utype> tensor, const utype scalar) {
  xcombos combos(tensor.shape, 'E');
  while (combos.increment()) {
    std::vector<int> indices = shiftIndices(combos.getCombo(), tensor.lindex);
    tensor(indices) = tensor(indices) * scalar;
  }
  return tensor;
}

template<class utype>
xtensor<utype> operator*(const utype scalar, const xtensor<utype>& tensor) {
  return tensor * scalar;
}

template<class utype>
xtensor<utype> operator/(xtensor<utype> tensor, const utype scalar) {
  xcombos combos(tensor.shape, 'E');
  while (combos.increment()) {
    std::vector<int> indices = shiftIndices(combos.getCombo(), tensor.lindex);
    tensor(indices) = (utype) tensor(indices)/scalar;
  }
  return tensor;
}
////// END with scalars

////// BEGIN with xtensors
template<class utype>
xtensor<utype> operator+(const xtensor<utype>& tensor1, const xtensor<utype>& tensor2) {
  bool LDEBUG = (false || XHOST.DEBUG || _DEBUG_XTENSOR_);
  if (LDEBUG) {
    std::cerr << _XTENSOR_ERR_PREFIX_ << "operator+ : Dimensions tensor 1";
    for (uint i = 0; i < tensor1.shape.size(); i++) {
      std::cerr << " " << tensor1.shape[i];
    }
    std::cerr << ", dimensions tensor 2:";
    for (uint i = 0; i < tensor2.shape.size(); i++) {
      std::cerr << " " << tensor2.shape[i];
    }
    std::cerr << std::endl;
  }
  xtensor<utype> tensor_sum(tensor1.shape);
  if (tensor1.sameShape(tensor2)) {
    xcombos combos(tensor1.shape, 'E');
    while (combos.increment()) {
      std::vector<int> indices = combos.getCombo();
      std::vector<int> indices1 = shiftIndices(indices, tensor1.lindex);
      std::vector<int> indices2 = shiftIndices(indices, tensor2.lindex);
      indices = shiftIndices(indices, 1);
      tensor_sum(indices) = (utype) tensor1(indices1) + (utype) tensor2(indices2);
    }
    return tensor_sum;
  // Exception handling
  } else {
    std::string function = _XTENSOR_ERR_PREFIX_ + "operator+";
    std::string message = "Tensors are of different size.";
    throw xerror(function, message, _INDEX_MISMATCH_);
  }
}

template<class utype>
xtensor<utype> operator-(const xtensor<utype>& tensor1, const xtensor<utype>& tensor2) {
  bool LDEBUG = (false || XHOST.DEBUG || _DEBUG_XTENSOR_);
  if (LDEBUG) {
    std::cerr << _XTENSOR_ERR_PREFIX_ << "operator- : Dimensions tensor 1";
    for (uint i = 0; i < tensor1.shape.size(); i++) {
      std::cerr << " " << tensor1.shape[i];
    }
    std::cerr << ", dimensions tensor 2:";
    for (uint i = 0; i < tensor2.shape.size(); i++) {
      std::cerr << " " << tensor2.shape[i];
    }
    std::cerr << std::endl;
  }
  xtensor<utype> tensor_sum(tensor1.shape);
  if (tensor1.sameShape(tensor2)) {
    xcombos combos(tensor1.shape, 'E');
    while (combos.increment()) {
      std::vector<int> indices = combos.getCombo();
      std::vector<int> indices1 = shiftIndices(indices, tensor1.lindex);
      std::vector<int> indices2 = shiftIndices(indices, tensor2.lindex);
      indices = shiftIndices(indices, 1);
      tensor_sum(indices) = (utype) tensor1(indices1) - (utype) tensor2(indices2);
    }
    return tensor_sum;
  // Exception handling
  } else {
    std::string function = _XTENSOR_ERR_PREFIX_ + "operator-";
    std::string message = "Tensors are of different size.";
    throw xerror(function, message, _INDEX_MISMATCH_);
  }
}

////// END with xtensors
////// BEGIN with _subtensors
template<class utype>
xtensor<utype> operator+(const xtensor<utype>& tensor, const _subtensor<utype>& st) {
  xtensor<utype> tensor_from_st(st);
  return tensor + tensor_from_st;
}

template<class utype>
xtensor<utype> operator+(const _subtensor<utype>& st, const xtensor<utype>& tensor) {
  return tensor + st;
}

template<class utype>
xtensor<utype> operator+(const _subtensor<utype>& st1, const _subtensor<utype>& st2) {
  xtensor<utype> tensor1(st1);
  xtensor<utype> tensor2(st2);
  return tensor1 + tensor2;
}

template<class utype>
xtensor<utype> operator-(const xtensor<utype>& tensor, const _subtensor<utype>& st) {
  xtensor<utype> tensor_from_st(st);
  return tensor - tensor_from_st;
}

template<class utype>
xtensor<utype> operator-(const _subtensor<utype>& st, const xtensor<utype>& tensor) {
  return -tensor + st;
}

template<class utype>
xtensor<utype> operator-(const _subtensor<utype>& st1, const _subtensor<utype>& st2) {
  xtensor<utype> tensor1(st1);
  xtensor<utype> tensor2(st2);
  return tensor1 - tensor2;
}

////// END with _subtensors
//// END binary operators
// END operators

// BEGIN member functions
//set/////////////////////////////////////////////////////////////////////////
template<class utype>
void xtensor<utype>::set(const utype& value) {
  for (int i = 0; i < nelements; i++) {
    corpus[i] = (utype) value;
  }
}

template<class utype>
void xtensor<utype>::reset() {
  for (int i = 0; i < nelements; i++) {
    corpus[i] = (utype) 0.0;
  }
}

template<class utype>
void xtensor<utype>::clear() {
  for (int i = 0; i < nelements; i++) {
    corpus[i] = (utype) 0.0;
  }
}

//END member functions

// BEGIN non-member functions
//// BEGIN conversion functions
////// From xtensor
template<class utype>
std::vector<utype> xtensor2vector(xtensor<utype>& tensor) {
  bool LDEBUG = (false || XHOST.DEBUG || _DEBUG_XTENSOR_);
  if (LDEBUG) {
    std::cerr << _XTENSOR_ERR_PREFIX_ << "xtensor2vector - Number of tensor dimensions: " << tensor.ndim << std::endl;
  }
  if (tensor.ndim > 1) {
    // Exception handling
    std::string function = _XTENSOR_ERR_PREFIX_ + "xtensor2vector";
    std::stringstream message;
    message << "Cannot convert xtensor with " << tensor.ndim << " dimensions to vector.";
    throw xerror(function, message, _INDEX_MISMATCH_);
  } else {
    int rows = tensor.shape[0];
    std::vector<utype> vec(rows);
    for (int i = 0; i < rows; i++) {
      vec[i] = tensor[i + tensor.lindex[0]];
    }
    return vec;
  }
}

template<class utype>
std::vector<utype> xtensor2vector(const _subtensor<utype>& st) {
  xtensor<utype> tensor(st);
  return xtensor2vector(tensor);
}

template<class utype>
xvector<utype> xtensor2xvector(xtensor<utype>& tensor) {
  bool LDEBUG = (false || XHOST.DEBUG || _DEBUG_XTENSOR_);
  if (LDEBUG) {
    std::cerr << _XTENSOR_ERR_PREFIX_ << "xtensor2xvector - Number of tensor dimensions: " << tensor.ndim << std::endl;
  }
  if (tensor.ndim > 1) {
    // Exception handling
    std::string function = _XTENSOR_ERR_PREFIX_ + "xtensor2xvector";
    std::stringstream message;
    message << "Cannot convert xtensor with " << tensor.ndim << " dimensions to xvector.";
    throw xerror(function, message, _INDEX_MISMATCH_);
  } else {
    int rows = tensor.shape[0];
    aurostd::xvector<utype> xvec(rows);
    for (int i = 0; i < rows; i++) {
      xvec(i + xvec.lrows) = tensor[i + tensor.lindex[0]];
    }
    return xvec;
  }
}

template<class utype>
xvector<utype> xtensor2xvector(const _subtensor<utype>& st) {
  xtensor<utype> tensor(st);
  return xtensor2xvector(tensor);
}

template<class utype>
xmatrix<utype> xtensor2xmatrix(xtensor<utype>& tensor) {
  bool LDEBUG = (false || XHOST.DEBUG || _DEBUG_XTENSOR_);
  if (LDEBUG) {
    std::cerr << _XTENSOR_ERR_PREFIX_ << "xtensor2xmatrix - Number of tensor dimensions: " << tensor.ndim << std::endl;
  }
  if (tensor.ndim > 2) {
    // Exception handling
    std::string function = _XTENSOR_ERR_PREFIX_ + "xtensor2xmatrix";
    std::stringstream message;
    message << "Cannot convert xtensor with " << tensor.ndim << " dimensions to xmatrix.";
    throw xerror(function, message, _INDEX_MISMATCH_);
  } else {
    int rows, cols;
    rows = tensor.shape[0];
    if (tensor.ndim > 1) {
      cols = tensor.shape[1];
    } else {
      cols = 1;
    }
    aurostd::xmatrix<utype> mat(rows, cols);
    for (int i = 0; i < rows; i++) {
      if (tensor.ndim > 1) {
        for (int j = 0; j < cols; j++) {
          mat(i + 1, j + 1) = tensor[i + tensor.lindex[0]][j + tensor.lindex[1]];
        }
      } else {
        mat(i, 1) = tensor[i+tensor.lindex[0]];
      }
    }
    return mat;
  }
}

template<class utype>
xmatrix<utype> xtensor2xmatrix(const _subtensor<utype>& st) {
  xtensor<utype> tensor(st);
  return xtensor2xmatrix(tensor);
}

////// To xtensor
template<class utype>
xtensor<utype> vector2xtensor(const std::vector<utype>& vec) {
  std::vector<int> dim(1, vec.size());
  xtensor<utype> tensor(dim);
  for (uint i = 0; i < vec.size(); i++) {
    tensor[i + 1] = vec[i];
  }
  return tensor;
}

template<class utype>
xtensor<utype> xvector2xtensor(const aurostd::xvector<utype>& xvec) {
  std::vector<int> dim(1, xvec.rows);
  xtensor<utype> tensor(dim);
  for (int i = 0; i < xvec.rows; i++) {
    tensor[i + 1] = xvec(xvec.lrows + i);
  }
  return tensor;
}

template<class utype>
xtensor<utype> xmatrix2xtensor(const aurostd::xmatrix<utype>& xmat) {
  std::vector<int> dim(2);
  dim[0] = xmat.rows; dim[1] = xmat.cols;
  xtensor<utype> tensor(dim);
  for (int i = 0; i < xmat.rows; i++) {
    for (int j = 0; j < xmat.cols; j++) {
      tensor[i + 1][j + 1] = xmat(xmat.lrows + i, xmat.lcols + j);
    }
  }
  return tensor;
}

//// END conversion functions

//// BEGIN return type xtensor
template<class utype>
xtensor<utype> abs(const xtensor<utype>& tensor_in) {
  xtensor<utype> tensor_out(tensor_in.uindex, tensor_in.lindex);
  xcombos combos(tensor_out.shape, 'E');
  while (combos.increment()) {
    std::vector<int> indices = shiftIndices(combos.getCombo(), tensor_in.lindex);
    tensor_out(indices) = aurostd::abs((utype) tensor_in(indices));
  }
  return tensor_out;
}

template<class utype>
xtensor<utype> abs(const _subtensor<utype>& st) {
  xtensor<utype> tensor_out(st);
  return abs(tensor_out);
}

template<class utype>
xtensor<utype> ceil(const xtensor<utype>& tensor_in) {
  xtensor<utype> tensor_out(tensor_in.uindex, tensor_in.lindex);
  xcombos combos(tensor_out.shape, 'E');
  while (combos.increment()) {
    std::vector<int> indices = shiftIndices(combos.getCombo(), tensor_in.lindex);
    tensor_out(indices) = std::ceil((utype) tensor_in(indices));
  }
  return tensor_out;
}

template<class utype>
xtensor<utype> ceil(const _subtensor<utype>& st) {
  xtensor<utype> tensor_out(st);
  return ceil(tensor_out);
}

template<class utype>
xtensor<utype> floor(const xtensor<utype>& tensor_in) {
  xtensor<utype> tensor_out(tensor_in.uindex, tensor_in.lindex);
  xcombos combos(tensor_out.shape, 'E');
  while (combos.increment()) {
    std::vector<int> indices = shiftIndices(combos.getCombo(), tensor_in.lindex);
    tensor_out(indices) = std::floor((utype) tensor_in(indices));
  }
  return tensor_out;
}

template<class utype>
xtensor<utype> floor(const _subtensor<utype>& st) {
  xtensor<utype> tensor_out(st);
  return floor(tensor_out);
}


template<class utype>
xtensor<utype> nint(const xtensor<utype>& tensor_in) {
  xtensor<utype> tensor_out(tensor_in.uindex, tensor_in.lindex);
  xcombos combos(tensor_out.shape, 'E');
  while (combos.increment()) {
    std::vector<int> indices = shiftIndices(combos.getCombo(), tensor_in.lindex);
    tensor_out(indices) = aurostd::nint((utype) tensor_in(indices));
  }
  return tensor_out;
}

template<class utype>
xtensor<utype> nint(const _subtensor<utype>& st) {
  xtensor<utype> tensor_out(st);
  return nint(tensor_out);
}

template<class utype>
xtensor<utype> round(const xtensor<utype>& tensor_in, const utype& tol) {
  xtensor<utype> tensor_out(tensor_in.uindex, tensor_in.lindex);
  xcombos combos(tensor_out.shape, 'E');
  while (combos.increment()) {
    std::vector<int> indices = shiftIndices(combos.getCombo(), tensor_in.lindex);
    tensor_out(indices) = roundoff((utype) tensor_in(indices), tol);
  }
  return tensor_out;
}

template<class utype>
xtensor<utype> round(const _subtensor<utype>& st, const utype& tol) {
  xtensor<utype> tensor_out(st);
  return round(tensor_out, tol);
}

template<class utype>
xtensor<utype> sign(const xtensor<utype>& tensor_in) {
  xtensor<utype> tensor_out(tensor_in.uindex, tensor_in.lindex);
  xcombos combos(tensor_out.shape, 'E');
  while (combos.increment()) {
    std::vector<int> indices = shiftIndices(combos.getCombo(), tensor_in.lindex);
    tensor_out(indices) = aurostd::sign((utype) tensor_in(indices));
  }
  return tensor_out;
}

template<class utype>
xtensor<utype> sign(const _subtensor<utype>& st) {
  xtensor<utype> tensor_out(st);
  return sign(tensor_out);
}

template<class utype>
 // Creates an identity tensor of rank n
xtensor<utype> identity_tensor(const utype& _type, int n) {  // _type is needed to deduce utype
  if (_type){;}  // to suppress compiler warninigs
  xtensor<utype> tensor(n);
  tensor.set((utype) 0.0); // In case the default initialization value is different from 0
  std::vector<int> diag(n);
  for (int i = 1; i <= n; i++) {
    for (int j = 0; j < n; j++) {
      diag[j] = i;
    }
    tensor(diag) = (utype) 1;
  }
  return tensor;
}
//// END return type xtensor

//// BEGIN return type utype
template<class utype>
utype max(const xtensor<utype>& tensor) {
  xcombos combos(tensor.shape, 'E');
  std::vector<int> indices = shiftIndices(combos.getCombo(), tensor.lindex);
  utype m = (utype) tensor(indices);
  while (combos.increment()) {
    std::vector<int> indices = shiftIndices(combos.getCombo(), tensor.lindex);
    if (tensor(indices) > m) {
      m = (utype) tensor(indices);
    }
  }
  return m;
}

template<class utype>
utype max(const _subtensor<utype>& st) {
  xtensor<utype> tensor_out(st);
  return max(tensor_out);
}

template<class utype>
utype min(const xtensor<utype>& tensor) {
  xcombos combos(tensor.shape, 'E');
  std::vector<int> indices = shiftIndices(combos.getCombo(), tensor.lindex);
  utype m = (utype) tensor(indices);
  while (combos.increment()) {
    std::vector<int> indices = shiftIndices(combos.getCombo(), tensor.lindex);
    if (tensor(indices) < m) {
      m = (utype) tensor(indices);
    }
  }
  return m;
}

template<class utype>
utype min(const _subtensor<utype>& st) {
  xtensor<utype> tensor_out(st);
  return min(tensor_out);
}

template<class utype>
utype sum(const xtensor<utype>& tensor) {
  utype s = (utype) 0;
  xcombos combos(tensor.shape, 'E');
  while (combos.increment()) {
    std::vector<int> indices = shiftIndices(combos.getCombo(), tensor.lindex);
    s += tensor(indices);
  }
  return s;
}

template<class utype>
utype sum(const _subtensor<utype>& st) {
  xtensor<utype> tensor_out(st);
  return sum(tensor_out);
}

template<class utype>
utype trace(xtensor<utype>& tensor) {
  bool LDEBUG = (false || XHOST.DEBUG || _DEBUG_XTENSOR_);
  if (LDEBUG) {
    std::cerr << _XTENSOR_ERR_PREFIX_ << "trace - Tensor shape:";
    for (uint i = 0; i < tensor.shape.size(); i++) {
      std::cerr << " " << tensor.shape[i];
    }
    std::cerr << std::endl;
  }
  utype trc = 0;
  if (tensor.is_cubic) {
    uint n = tensor.ndim;
    std::vector<int> diag(n);
    for (uint i = 0; i < n; i++) {
      for (uint j = 0; j < n; j++) {
        diag[j] = i + tensor.lindex[j];
      }
      trc += tensor(diag);
    }
    return trc;
  } else {
    std::string function = _XTENSOR_ERR_PREFIX_ + "trace";
    std::string message = "Trace is only defined for cubic tensors.";
    throw xerror(function, message, _RUNTIME_ERROR_);
  }
}

template<class utype>
utype trace(const _subtensor<utype>& st) {
  xtensor<utype> tensor_out(st);
  return trace(tensor_out);
}


//// END return type utype
// END non-member functions

}  // namespace aurostd
/******************************** END XTENSOR *******************************/

/****************************** OTHER FUNCTIONS *****************************/
namespace aurostd {
std::vector<int> shiftIndices(std::vector<int> vec, const std::vector<int>& shiftvec) {
  if (vec.size() == shiftvec.size()) {
    for (uint i = 0; i < vec.size(); i++) {
      vec[i] += shiftvec[i];
    }
  } else {
    std::string function = "aurostd::shiftIndices";
    std::string message = "Vectors have different sizes.";
    throw xerror(function, message, _INDEX_MISMATCH_);
  }
  return vec;
}

std::vector<int> shiftIndices(std::vector<int> vec, const int& shift) {
  for (uint i = 0; i < vec.size(); i++) {
    vec[i] += shift;
  }
  return vec;
}
} // aurostd
/****************************** OTHER FUNCTIONS *****************************/

/*********************************** EIJK ***********************************/
// eijk (Ricci Tensor) from old xtensor - used in aflow_xatom.cpp
namespace aurostd {
  int eijk(int i, int j, int k) {
    int ii = (i - 1) % 3 + 1;
    int jj = (j - 1) % 3 + 1;
    int kk = (k - 1) % 3 + 1;
    if ((ii == 1) && (jj == 2) && (kk == 3)) return 1;
    if ((ii == 2) && (jj == 3) && (kk == 1)) return 1;
    if ((ii == 3) && (jj == 1) && (kk == 2)) return 1;
    if ((ii == 1) && (jj == 3) && (kk == 2)) return -1;
    if ((ii == 3) && (jj == 2) && (kk == 1)) return -1;
    if ((ii == 2) && (jj == 1) && (kk == 3)) return -1;
    return 0;
  }

  int eijk(xvector<int> ijk) {
    return eijk(ijk[1], ijk[2], ijk[3]);
  }

  int estarijk(int i, int j, int k) {
    int ii = (i - 1) % 3 + 1;
    int jj = (j - 1) % 3 + 1;
    int kk = (k - 1) % 3 + 1;
    if ((ii == 1) && (jj == 2) && (kk == 3)) return 1;
    if ((ii == 2) && (jj == 3) && (kk == 1)) return 1;
    if ((ii == 3) && (jj == 1) && (kk == 2)) return 1;
    return 0;
  }

  int estarijk(xvector<int> ijk) {
    return estarijk(ijk[1], ijk[2], ijk[3]);
  }
} // namespace aurostd
/*********************************** EIJK ***********************************/

#endif
//****************************************************************************
// *                                                                         *
// *           Aflow STEFANO CURTAROLO - Duke University 2003-2018           *
// *                  Marco Esters - Duke University 2018                    *
// *                                                                         *
//****************************************************************************
